#!/usr/bin/env python3
"""Run and view a kMC model. For this to work one needs a
kmc_model.(so/pyd) and a kmc_settings.py in the import path."""
from __future__ import print_function
#    Copyright 2009-2013 Max J. Hoffmann (mjhoffmann@gmail.com)
#    This file is part of kmcos.
#
#    kmcos is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    kmcos is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with kmcos.  If not, see <http://www.gnu.org/licenses/>.

import multiprocessing
import threading
import os
import sys

import numpy as np
import time


from ase.gui.images import Images
import ase

try:
    import gtk
    import gobject
except Exception as e:
    try:
        import gi
        gi.require_version('Gtk', '3.0')
        from gi.repository import Gtk as gtk
        from gi.repository import GObject as gobject
        from ase.gui.view import View
        from ase.gui.status import Status
        from ase.gui.gui import GUI
    except Exception as e:
        View = type('View', (), {})
        Status = type('Status', (), {})
        print('Warning: GTK not available. Cannot run graphical front-end')
        print(e)
try:
    import matplotlib
    gtk_version = 'GTKAgg' #initailizing
    if sys.version_info.major == 3:
      gtk_version = 'GTK3Agg'
    if os.name == 'posix':
        matplotlib.use(gtk_version)
    elif os.name == 'nt':
        matplotlib.use('wxagg')
    else:
        matplotlib.use(gtk_version)
    import matplotlib.pylab as plt
except Exception as e:
    print('Could not import matplotlib frontend for real-time plotting')
    print(e)

from kmcos.run import KMC_Model, get_tof_names, lattice, settings


class ParamSlider(gtk.HScale):
    """A horizontal slider bar allowing the user to adjust a model parameter
    at runtime.
    """

    def __init__(self, name, value, xmin, xmax, scaleInput, parameter_callback):
        super().__init__()
        self.parameter_callback = parameter_callback
        self.resolution = 1000.
        adjustment = gtk.Adjustment(0, 0, self.resolution, 0.1, 1.)
        self.xmin = float(xmin)
        self.xmax = float(xmax)
        if self.xmin == self.xmax:
            self.xmax = self.xmax + 1.
        self.settings = settings
        self.param_name = name
        self.scaleA = scaleInput
        gtk.HScale.__init__(adjustment)
        
        self.connect('format-value', self.linlog_scale_format)
        self.connect('value-changed', self.value_changed)
        self.set_tooltip_text(self.param_name)
        if self.scaleA == 'linear':
            scaled_value = (self.resolution * (float(value) - self.xmin) /
                                               (self.xmax - self.xmin))
            self.set_value(scaled_value)
        elif self.scaleA == 'log':
            scaled_value = 1000 * (np.log(float(value) / self.xmin) /
                                   np.log(float(self.xmax / self.xmin)))
            self.set_value(scaled_value)

    def linlog_scale_format(self, _widget, value):
        """Format a model parameter's name for display above
        the slider bar.
        """
        value /= self.resolution
        name = self.param_name
        unit = ''
        if self.param_name.endswith('gas'):
            name = name[:-3]
        if self.param_name.startswith('p_'):
            name = 'p(%s)' % name[2:]
            unit = 'bar'
        if name == 'T':
            unit = 'K'
        if self.scaleA == 'log':
            vstr = '%s: %.2e %s (log)' % (name,
                           self.xmin * (self.xmax / self.xmin) ** value, unit)
        elif self.scaleA == 'linear':
            vstr = '%s: %s %s' % (name,
                           self.xmin + value * (self.xmax - self.xmin), unit)
        else:
            raise UserWarning("Unexpected scale mode %s" % self.scaleA)
        return vstr

    def value_changed(self, _widget):
        """Handle the event, that slider bar has been dragged."""
        scale_value = self.get_value() / self.resolution
        if self.scaleA == 'log':
            value = self.xmin * (self.xmax / self.xmin) ** scale_value
        else:
            value = self.xmin + (self.xmax - self.xmin) * scale_value
        self.parameter_callback(self.param_name, value)


class FakeWidget():
    """This widget is used by FakeUI containing the menu
    base settings that the ase.gui.images modules expects.
    """

    def __init__(self, path):
        self.active = False
        if path.endswith('ShowUnitCell'):
            self.active = True
        elif path.endswith('ShowBonds'):
            self.active = False
        elif path.endswith('ShowAxes'):
            self.active = True

    def get_active(self):
        return self.active


class FakeUI():
    """This is a fudge class to simulate to the View class a non-existing
    menu with included settings
    """

    def __init__(self):
        return self

    def get_widget(self, path):
        widget = FakeWidget(path)
        return widget


class KMC_ViewBox(threading.Thread, View, Status, FakeUI):
    """The part of the viewer GUI that displays the model's
    current configuration.
    """

    def __init__(self, queue, signal_queue, vbox, window,
                rotations='', show_unit_cell=True, show_bonds=False):

        threading.Thread.__init__(self)
        self.image_queue = queue
        self.signal_queue = signal_queue
        self.configured = False
        self.ui = FakeUI.__init__(self)
        self.images = Images()
        self.aseGui = GUI()
        #self.aseGui.images.initialize([ase.atoms.Atoms()])
        self.images.initialize([ase.atoms.Atoms()])
        self.killed = False
        self.paused = False

        self.vbox = vbox
        self.window = window

        self.vbox.connect('scroll-event', self.scroll_event)
        self.window.connect('key-press-event', self.on_key_press)
        rotations = '0.0x,0.0y,0.0z'#[3,3,3]#np.zeros(3)
        self.config = {'force_vector_scale': None, 'velocity_vector_scale': None, 'swap_mouse': False}
        View.__init__(self, rotations)
        Status.__init__(self)
        self.vbox.show()

        if os.name == 'posix':
            self.live_plot = True
        else:
            self.live_plot = False

        #self.drawing_area.realize()
        self.scaleA = 3.0
        self.center = np.array([8, 8, 8])
        #self.set_colors()
        #self.set_coordinates(0)
        self.center = np.array([0, 0, 0])

        self.tofs = get_tof_names()

        # history tracking arrays
        self.times = []
        self.tof_hist = []
        self.occupation_hist = []

        # prepare diagrams
        self.data_plot = plt.figure()
        #plt.xlabel('$t$ in s')
        self.tof_diagram = self.data_plot.add_subplot(211)
        self.tof_diagram.set_yscale('log')
        #self.tof_diagram.get_yaxis().get_major_formatter().set_powerlimits(
                                                                    #(3, 3))
        self.tof_plots = []
        for tof in self.tofs:
            self.tof_plots.append(self.tof_diagram.plot([], [], label=tof)[0])

        self.tof_diagram.legend(loc='lower left')
        self.tof_diagram.set_ylabel(
            'TOF in $\mathrm{s}^{-1}\mathrm{site}^{-1}$')
        self.occupation_plots = []
        self.occupation_diagram = self.data_plot.add_subplot(212)
        for species in sorted(settings.representations):
            self.occupation_plots.append(
                self.occupation_diagram.plot([], [], label=species)[0],)
        self.occupation_diagram.legend(loc=2)
        self.occupation_diagram.set_xlabel('$t$ in s')
        self.occupation_diagram.set_ylabel('Coverage')

    def update_vbox(self, atoms):
        """Update the ViewBox."""
        if not self.center.any():
            self.center = atoms.cell.diagonal() * .5
        #self.images = Images([atoms])
        #self.images.filenames = ['kmcos GUI - %s' % settings.model_name]
        #self.set_colors()
        #self.set_coordinates(0)
        #self.set_atoms(atoms)
        self.scale = self.scaleA
        self.aseGui.scale = self.scale
        atoms.center(vacuum=3.0)
        self.aseGui.images.initialize([atoms])
        self.aseGui.images.center()
        self.aseGui.set_frame()
        #self.draw()
        #self.label.set_label('%.3e s (%.3e steps)' % (atoms.kmc_time,
        #                                            atoms.kmc_step))

    def update_plots(self, atoms):
        """Update the coverage and TOF plots."""
        # fetch data piggy-backed on atoms object
        new_time = atoms.kmc_time

        occupations = atoms.occupation.sum(axis=1) / lattice.spuck
        tof_data = atoms.tof_data

        # store locally
        while len(self.times) > getattr(settings, 'hist_length', 30):
            self.tof_hist.pop(0)
            self.times.pop(0)
            self.occupation_hist.pop(0)

        self.times.append(atoms.kmc_time)
        self.tof_hist.append(tof_data)
        self.occupation_hist.append(occupations)

        # plot TOFs
        for i, tof_plot in enumerate(self.tof_plots):
            tof_plot.set_xdata(self.times)
            tof_plot.set_ydata([tof[i] for tof in self.tof_hist])
            self.tof_diagram.set_xlim(self.times[0], self.times[-1])
            self.tof_diagram.set_ylim(1e-3,
                      10 * max([tof[i] for tof in self.tof_hist]))

        # plot occupation
        for i, occupation_plot in enumerate(self.occupation_plots):
            occupation_plot.set_xdata(self.times)
            occupation_plot.set_ydata(
                            [occ[i] for occ in self.occupation_hist])
        max_occ = max(occ[i] for occ in self.occupation_hist)
        self.occupation_diagram.set_ylim([0, max(1, max_occ)])
        self.occupation_diagram.set_xlim([self.times[0], self.times[-1]])

        self.data_plot.canvas.draw_idle()
        manager = plt.get_current_fig_manager()
        if hasattr(manager, 'toolbar'):
            toolbar = manager.toolbar
            if hasattr(toolbar, 'set_visible'):
                toolbar.set_visible(False)

        plt.show()

        # [:] is necessary so that it copies the
        # values and doesn't reinitialize the pointer
        self.time = new_time

        return False

    def kill(self):
        self.killed = True

    def run(self):
        time.sleep(1.)
        while not self.killed:
            time.sleep(1)
            if not self.image_queue.empty():
                atoms = self.image_queue.get()
                gobject.idle_add(self.update_vbox, atoms)
                if self.live_plot:
                    gobject.idle_add(self.update_plots, atoms)

    def on_key_press(self, _widget, event):
        """Process key press event on view box."""
        signal_dict = {'a': 'ACCUM_RATE_SUMMATION',
                       'c': 'COVERAGE',
                       'd': 'DOUBLE',
                       'h': 'HALVE',
                       's': 'SWITCH_SURFACE_PROCESSS_OFF',
                       'S': 'SWITCH_SURFACE_PROCESSS_ON',
                       'w': 'WRITEOUT',
                      }
        if event.string in [' ', 'p']:
            if not self.paused:
                self.signal_queue.put('PAUSE')
                self.paused = True
            else:
                self.signal_queue.put('START')
                self.paused = False
        elif event.string in ['?']:
            for key, command in list(signal_dict.items()):
                print('%4s %s' % (key, command))
        elif event.string in signal_dict:
            self.signal_queue.put(signal_dict.get(event.string, ''))

    def scroll_event(self, _window, event):
        """Zoom in/out when using mouse wheel"""
        x = 1.0
        if event.direction == gtk.gdk.SCROLL_UP:
            x = 1.2
        elif event.direction == gtk.gdk.SCROLL_DOWN:
            x = 1.0 / 1.2
        self._do_zoom(x)

    def _do_zoom(self, x):
        """Utility method for zooming"""
        self.scaleA *= x
        try:
            atoms = self.image_queue.get()
        except Exception as e:
            atoms = ase.atoms.Atoms()
            print(e)
        self.update_vbox(atoms)


class KMC_ModelProxy(multiprocessing.Process):
    """This is a proxy class handling the communication
    with the Model process.

    This proxy was necessary because Windows does not
    support fork() and thus the model process cannot be split-off
    directly. As a workaround multiprocessing tries to pickle the
    memory of the current process, however it does not know how
    to pickle the fortran objects.
    """
    def __init__(self, *args, **kwargs):
        super(KMC_ModelProxy, self).__init__()
        self.steps_per_frame = kwargs.get('steps_per_frame', 50000)
        self.model = kwargs.get('model', None, )
        self.kwargs = kwargs
        self.signal_queue = self.kwargs.get('signal_queue')
        self.parameter_queue = self.kwargs.get('parameter_queue')
        self.queue = self.kwargs.get('queue')

    def run(self):
        if self.model is None:
            self.model = KMC_Model(self.queue,
                                   self.parameter_queue,
                                   self.signal_queue,
                                   steps_per_frame=self.steps_per_frame)
        self.model.run()

    def join(self):
        self.signal_queue.put('JOIN')
        super(KMC_ModelProxy, self).join()

    def terminate(self):
        self.signal_queue.put('STOP')
        super(KMC_ModelProxy, self).terminate()


class KMC_Viewer():
    """A graphical front-end to run, manipulate
    and view a kMC model.
    """

    def __init__(self, model=None, steps_per_frame=50000):
        if sys.version_info.major == 3:
            self.window = gtk.Window()
        else:
            self.window = gtk.Window(gtk.WINDOW_TOPLEVEL)
            self.window.set_position(gtk.WIN_POS_CENTER)
            
        self.window.connect('delete-event', self.exit)

        self.vbox = gtk.VBox()
        self.window.add(self.vbox)
        queue = multiprocessing.Queue(maxsize=3)
        self.parameter_queue = multiprocessing.Queue(maxsize=50)
        self.signal_queue = multiprocessing.Queue(maxsize=10)
        if model is None:
            self.model = KMC_ModelProxy(queue=queue,
                                        parameter_queue=self.parameter_queue,
                                        signal_queue=self.signal_queue,
                                        steps_per_frame=steps_per_frame)
        else:
            self.model = model
            self.model.image_queue = queue
            self.model.parameter_queue = self.parameter_queue
            self.model.signal_queue = self.signal_queue
        self.viewbox = KMC_ViewBox(queue, self.signal_queue,
                                   self.vbox, self.window)

        adjustable_params = [param for param in settings.parameters
                             if settings.parameters[param]['adjustable']]

        for param_name in sorted(adjustable_params):
            param = settings.parameters[param_name]
            slider = ParamSlider(param_name, param['value'],
                                 param['min'], param['max'],
                                 param['scale'], self.parameter_callback)
            self.vbox.add(slider)
            self.vbox.pack_start(self.vbox, False, False, 0)
            start_pack = self.vbox.query_child_packing(slider).pack_type
            self.vbox.set_child_packing(slider, expand=False,
                                        fill=False, padding=0,
                                        pack_type=start_pack)
        self.window.set_title('kmcos GUI')
        self.window.show_all()

    def parameter_callback(self, name, value):
        """Sent (updated) parameters to the model process."""
        settings.parameters[name]['value'] = value
        self.parameter_queue.put(settings.parameters)

    def exit(self, _widget, _event):
        """Exit the viewer application cleanly
        killing all subprocesses before the main
        process.
        """

        self.viewbox.kill()
        #print(' ... sent kill to viewbox')
        self.viewbox.join()
        #print(' ... viewbox thread joined')
        self.signal_queue.put('STOP')
        #print(' ... sent stop to model')
        self.model.terminate()
        self.model.join()
        #print(' ... model thread joined')
        #base.deallocate_system()
        gtk.main_quit()
        return True


def main(model=None, steps_per_frame=50000):
    """The entry point for the kmcos viewer application. In order to
    run and view a model the corresponding kmc_settings.py and
    kmc_model.(so/pyd) must be in the current import path, e.g. ::

        from sys import path
        path.append('/path/to/model')
        from kmcos.view import main
        main() # launch viewer
    """

    gobject.threads_init()
    viewer = KMC_Viewer(model, steps_per_frame=steps_per_frame)
    viewer.model.start()
    viewer.viewbox.start()
    gtk.main()



#!/usr/bin/env python3
# coding: utf-8
"""
## Initial Bias Detection

In the following we will run some numerical experiments using the
initial bias detection proposed by M. Rossetti (Univ. Arkansas)
[link](http://dx.doi.org/10.1109/WSC.2005.1574321)

Helpful literature:
“Wiley: Introduction to Statistical Quality Control, 7th Edition - Douglas C. Montgomery.” Accessed March 8, 2016. http://www.wiley.com/WileyCDA/WileyTitle/productCd-EHEP002023.html.
Hoad, Kathryn, Stewart Robinson, and Ruth Davies. “Automating Warm-up Length Estimation.” In Proceedings of the 40th Conference on Winter Simulation, 532–40. WSC ’08. Miami, Florida: Winter Simulation Conference, 2008. http://dl.acm.org/citation.cfm?id=1516744.1516846.
Rossetti, M. D., Zhe Li, and Peng Qu. “Exploring Exponentially Weighted Moving Average Control Charts to Determine the Warm-up Period.” In Simulation Conference, 2005 Proceedings of the Winter, 10 pp. – , 2005. doi:10.1109/WSC.2005.1574321.

"""
from __future__ import print_function

# import the necessary python modules
import numpy as np
import pprint
import itertools


# Calculate the LCL, and UCL according to expontially weighted moving
# average (EWMA)

def ewma_alpha(y, alpha, prev_ewma=None, adjust=True):
    """Adapted from pandas.algos.ewma

    """
    old_wt_factor = 1. - alpha
    old_wt = 1.
    new_wt = 1. if adjust else alpha

    alpha = np.double(alpha)
    if prev_ewma is not None:
        return alpha * y[-1] + (1 - alpha) * prev_ewma
    else:
        ewma = np.zeros_like(y, dtype=np.double)
        ewma[0] = y[0]
        # for i in range(len(y) - 1, -1, -1):
        for i in range(1, len(y)):
            old_wt *= old_wt_factor
            ewma[i] = (new_wt * y[i] + old_wt * ewma[i - 1]) / \
                (old_wt + new_wt)
            if adjust:
                old_wt += new_wt
            else:
                old_wt = 1.
        return ewma


def lcl_ucl(y, cutoff, L, lambda_factor):
    """Calculate the lower and upper control limit.

    cf. Rossetti, M. D., Zhe Li, and Peng Qu. “Exploring Exponentially Weighted Moving Average Control Charts to Determine the Warm-up Period.” In Simulation Conference, 2005 Proceedings of the Winter, 10 pp. – , 2005. doi:10.1109/WSC.2005.1574321.
    cf. “Wiley: Introduction to Statistical Quality Control, 7th Edition - Douglas C. Montgomery.” Accessed March 8, 2016. http://www.wiley.com/WileyCDA/WileyTitle/productCd-EHEP002023.html.

    """

    mu0 = np.mean(y[cutoff:]) * np.ones_like(y)
    sigma = np.std(y[cutoff:])
    N = len(y)
    delta = L * sigma * \
        np.sqrt(lambda_factor / (2 - lambda_factor)
                * (1 - (1 - lambda_factor)**np.arange(0., N, 1)))
    #print(N, d)
    return mu0, mu0 - delta, mu0 + delta


def p2d(y, cutoff, L, alpha):
    """Define the control function: this function estimates what how many
    points are under control if the corresponding EWMA falls into the LCL
    and UCL window.

    cf. Rossetti, M. D., Zhe Li, and Peng Qu. “Exploring Exponentially Weighted Moving Average Control Charts to Determine the Warm-up Period.” In Simulation Conference, 2005 Proceedings of the Winter, 10 pp. – , 2005. doi:10.1109/WSC.2005.1574321.
    """

    ewma = ewma_alpha(y, alpha)
    _, lcl, ucl = lcl_ucl(y, cutoff, L, alpha)
    return ((lcl[cutoff:] < ewma[cutoff:]) & (ewma[cutoff:] < ucl[cutoff:])).sum() / (float(len(y) - cutoff))


def get_scrap_fraction(y, L, alpha, warm_up):
    """Calculate which part of the data should be scrapped because it does not yet appear
    to be in steady-state.

    """

    if (y[0] == y).all():
        return 0.
    D = np.array([p2d(y, cutoff, L, alpha) for cutoff in range(len(y))])
    return (np.argmax(D[warm_up:]) + warm_up) / float(len(y))


def plot_normal(y, n=-1, *args, **kwargs):
    """Plot normalized data with the nth value
    """
    from matplotlib import pyplot as plt
    plt.plot(y / y[n], *args, **kwargs)


def make_ewma_plots(data, L, alpha, bias_threshold, seed):
    """Make so-called EWMA plots.

    “Wiley: Introduction to Statistical Quality Control, 7th Edition - Douglas C. Montgomery.” Accessed March 8, 2016. http://www.wiley.com/WileyCDA/WileyTitle/productCd-EHEP002023.html.

    Most for debugging purposes if the EWMA based steady-state analysis makes was sensible.
    """
    from matplotlib import pyplot as plt
    for key, y in list(data.items()):
        if not 'time' in key or 'step' in key:
            y = np.array(y)
            plt.clf()
            cutoff0 = int(bias_threshold * len(y))
            mu0, lcl, ucl = lcl_ucl(y, cutoff0, L, alpha)
            D = np.array([p2d(y, cutoff, L, alpha) for cutoff in range(len(y))])
            plot_normal(ewma_alpha(y, alpha), label='EWMA')
            plt.plot(y / ewma_alpha(y, alpha), label="signal")
            plot_normal(mu0, label='mean'.format(**locals()))
            plot_normal(lcl, n=0, label='LCL@{cutoff0}'.format(**locals()))
            plot_normal(ucl, n=0, label='UCL@{cutoff0}'.format(**locals()))
            plt.plot(D, 'k-', label='p2d')
            plt.text(np.argmax(D), .5, str(np.argmax(D)))
            legend = plt.legend()
            legend.get_frame().set_alpha(0.5)
            plt.savefig("{seed}_{key}.png".format(**locals()))


def sample_steady_state(model, batch_size=1000000,
                        L=4,
                        alpha=0.05,
                        bias_threshold=0.15,
                        tof_method='integ',
                        warm_up=20,
                        check_frequency=10,
                        show_progress=True,
                        make_plots=False,
                        output='str',
                        seed='EWMA',):
    """
        Run kMC model and continuously deploy steady-state detection to ensure that an initial bias does not enter the data.
        The steady-state detection is based on
        Rossetti, M. D., Zhe Li, and Peng Qu.
        "Exploring Exponentially Weighted Moving Average Control Charts to Determine the Warm-up Period."
        In Simulation Conference, 2005 Proceedings of the Winter, 10
        pp. – , 2005. doi:10.1109/WSC.2005.1574321.

        Define $L$  and $\alpha$ as in source as method parameters. Source
        suggesst ($L=3$, $\alpha=0.05$). Thorough tests showed that for $L=3$
        can fail to give accurate estimate if the statistical noise on the
        function is extremely small ($0<0.01$) because the $3\sigma$ environment
        was estimated too narrow. It was therefore decided that $L=4$ give more
        robust outcomes.

        Arguments:

            :param batch_size: This is the number of kMC steps per sampling batch. The number should be equal of larger to an average auto-correlation length of the kMC trajectory. A larger batch-size is inefficient, the shorter batch-size will confuse the exponentially weighted moving average (EWMA). If the progress is turned on and the progress-bar just jumps erratically even after 100s of batches, the number should be increased. The ideal batch-size can also depend on simulations and generally grow larger around phase-transitions.
            :type batch_size: int
            :param L: This sets the confidence range of the lower control limit (LCL) and upper control limit (UCL) of the EWMA. The source recommends a value of $L=3$. Test have shown that $L=4$ leads to more stable outcomes when noise is very small.
            :type L: int
            :param alpha: The scaling factor of the EWMA. Values between 0.01 and 0.1 have shown to give good performance.
            :type alpha: float
            :param show_progress: Show status of convergence in ASCII status bar (default: True)
            :type show_progress: bool
            :param make_plots: If True the steady-state estimator will make EWMA plots at every 100 steps as well as after convergence (default: False)
            :type make_plots: bool
            :param seed: Prefix string for EWMA plots (default: EWMA)
            :type seed: str
            :param tof_method: Forward of the same named option in get_std_sampled_data. Choose 'integ' to calculate rate based on coverages, choose 'procrates' to calculate rates based on actual events.
            :type tof_method: str
            :param warm_up: Number of batches to run before checking for steady-state (default: 20). When you begin checking to early, result usually have little statistical bearing and may mostly likely lead to erroneous results.
            :type warm_up: int
            :param check_frequency: Number specifying after every how many batches we check for steady-state. This is to reduce the computational cost of checking for steady-state. (default: 20).
            :type check_frequency: int
    """
    hist = {}

    if show_progress:
        import kmcos.utils.progressbar
        progress_bar = kmcos.utils.progressbar.ProgressBar()

    for batch in itertools.count():
        data = model.get_std_sampled_data(
            100, batch_size, tof_method=tof_method, output='dict')

        for key, data_point in list(data.items()):
            hist.setdefault(key, []).append(data_point)

        max_scrap = 0.
        critical_key = ''
        if batch < warm_up:
            if show_progress:
                progress_bar.render(
                    int(0), "Warm-up phase {batch}/{warm_up}".format(**locals()))

        else:
            if batch % check_frequency == 0:
                for key, y in list(hist.items()):
                    if 'time' in key or 'step' in key:
                        continue
                    scrap_fraction = get_scrap_fraction(
                        np.array(y), L, alpha, warm_up)
                    if scrap_fraction > max_scrap:
                        max_scrap = scrap_fraction
                        critical_key = key
                completed_percent = float(
                    1 - max_scrap) / (1 - bias_threshold) * 100.

                if make_plots and batch % 100 == 0:
                    make_ewma_plots(
                        hist, L, alpha, bias_threshold, seed="{seed}_{batch}".format(**locals()))
                if show_progress:
                    progress_bar.render(int(
                        completed_percent), "Limited by {critical_key:40s} ({batch})".format(**locals()))

                if completed_percent >= 100 and batch >= warm_up:
                    if show_progress:
                        print("Done after {batch} batches!".format(**locals()))
                    if make_plots:
                        make_ewma_plots(
                            hist, L, alpha, bias_threshold, seed="{seed}_final".format(**locals()))
                    break

    steady_state_start = int(batch * bias_threshold)
    for key, value in list(hist.items()):
        hist[key] = np.array(value[steady_state_start:])

    data = {}
    for key, values in list(hist.items()):
        if 'time' in key:
            data[key] = values[-1]
        elif 'step' in key:
            data[key] = sum(values)
        else:
            data[key] = np.average(values, weights=hist['kmc_time'])
    if output == 'dict':
        return data
    elif output == 'str':
        return ' '.join(format(data[key.replace('#', '')], '.5e') for key in model.get_std_header().split()) + '\n'

if __name__ == '__main__':
    import kmcos.run
    with kmcos.run.KMC_Model(banner=False, print_rates=False) as model:
        hist = sample_steady_state(
            model, 100000, tof_method='integ', show_progress=True, make_plots=True)
    print(model.get_std_header())
    print(hist)

